# type: ignore
"""
* | File        :	  epd7in5.py
* | Author      :   Waveshare team
* | Function    :   Electronic paper driver
* | Info        :
*----------------
* | This version:   V4.0
* | Date        :   2019-06-20
# | Info        :   python demo
-----------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to  whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS OR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
# pylint: disable=missing-function-docstring,missing-class-docstring,no-member

from logging import debug

from wg_utilities.devices.epd import epdconfig

# Display resolution
EPD_WIDTH = 800
EPD_HEIGHT = 480


# noinspection PyUnresolvedReferences,PyMissingOrEmptyDocstring,SpellCheckingInspection
class EPD:
    def __init__(self):
        self.reset_pin = epdconfig.RST_PIN
        self.dc_pin = epdconfig.DC_PIN
        self.busy_pin = epdconfig.BUSY_PIN
        self.cs_pin = epdconfig.CS_PIN
        self.width = EPD_WIDTH
        self.height = EPD_HEIGHT

    # Hardware reset
    def reset(self):
        epdconfig.digital_write(self.reset_pin, 1)
        epdconfig.delay_ms(200)
        epdconfig.digital_write(self.reset_pin, 0)
        epdconfig.delay_ms(2)
        epdconfig.digital_write(self.reset_pin, 1)
        epdconfig.delay_ms(200)

    def send_command(self, command):
        epdconfig.digital_write(self.dc_pin, 0)
        epdconfig.digital_write(self.cs_pin, 0)
        epdconfig.spi_writebyte([command])
        epdconfig.digital_write(self.cs_pin, 1)

    def send_data(self, data):
        epdconfig.digital_write(self.dc_pin, 1)
        epdconfig.digital_write(self.cs_pin, 0)
        epdconfig.spi_writebyte([data])
        epdconfig.digital_write(self.cs_pin, 1)

    def read_busy(self):
        debug("e-Paper busy")
        self.send_command(0x71)
        busy = epdconfig.digital_read(self.busy_pin)
        while busy == 0:
            self.send_command(0x71)
            busy = epdconfig.digital_read(self.busy_pin)
        epdconfig.delay_ms(200)

    def init(self):
        if epdconfig.module_init() != 0:
            return -1
        # EPD hardware init start
        self.reset()

        self.send_command(0x01)  # POWER SETTING
        self.send_data(0x07)
        self.send_data(0x07)  # VGH=20V,VGL=-20V
        self.send_data(0x3F)  # VDH=15V
        self.send_data(0x3F)  # VDL=-15V

        self.send_command(0x04)  # POWER ON
        epdconfig.delay_ms(100)
        self.read_busy()

        self.send_command(0x00)  # PANNEL SETTING
        self.send_data(0x1F)  # KW-3f   KWR-2F	BWROTP 0f	BWOTP 1f

        self.send_command(0x61)  # tres
        self.send_data(0x03)  # source 800
        self.send_data(0x20)
        self.send_data(0x01)  # gate 480
        self.send_data(0xE0)

        self.send_command(0x15)
        self.send_data(0x00)

        self.send_command(0x50)  # VCOM AND DATA INTERVAL SETTING
        self.send_data(0x10)
        self.send_data(0x07)

        self.send_command(0x60)  # TCON SETTING
        self.send_data(0x22)

        # EPD hardware init end
        return 0

    def getbuffer(self, image):
        # logging.debug("bufsiz = ",int(self.width/8) * self.height)
        buf = [0xFF] * (int(self.width / 8) * self.height)
        image_monocolor = image.convert("1")
        imwidth, imheight = image_monocolor.size
        pixels = image_monocolor.load()
        # logging.debug("imwidth = %d, imheight = %d",imwidth,imheight)
        if imwidth == self.width and imheight == self.height:
            debug("Vertical")
            for y in range(imheight):
                for x in range(imwidth):
                    # Set the bits for the column of pixels at the current position.
                    if pixels[x, y] == 0:
                        buf[int((x + y * self.width) / 8)] &= ~(0x80 >> (x % 8))
        elif imwidth == self.height and imheight == self.width:
            debug("Horizontal")
            for y in range(imheight):
                for x in range(imwidth):
                    new_x = y
                    new_y = self.height - x - 1
                    if pixels[x, y] == 0:
                        buf[int((new_x + new_y * self.width) / 8)] &= ~(0x80 >> (y % 8))
        return buf

    def display(self, image):
        self.send_command(0x13)
        for i in range(0, int(self.width * self.height / 8)):
            self.send_data(~image[i])

        self.send_command(0x12)
        epdconfig.delay_ms(100)
        self.read_busy()

    def clear(self):
        self.send_command(0x10)
        for _ in range(0, int(self.width * self.height / 8)):
            self.send_data(0x00)

        self.send_command(0x13)
        for _ in range(0, int(self.width * self.height / 8)):
            self.send_data(0x00)

        self.send_command(0x12)
        epdconfig.delay_ms(100)
        self.read_busy()

    def sleep(self):
        self.send_command(0x02)  # POWER_OFF
        self.read_busy()

        self.send_command(0x07)  # DEEP_SLEEP
        self.send_data(0xA5)

    @staticmethod
    def dev_exit():
        epdconfig.module_exit()



"""Custom client for interacting with Google's Photos API"""
from __future__ import annotations

from datetime import datetime
from enum import Enum, auto
from json import dumps
from logging import Logger
from os import getenv
from os.path import isfile, join
from typing import Iterable, TypedDict, cast

from requests import get

from wg_utilities.clients._google import GoogleClient, _GoogleEntityInfo
from wg_utilities.functions import force_mkdir, user_data_dir

LOCAL_MEDIA_DIRECTORY = getenv(
    "LOCAL_MEDIA_DIRECTORY", user_data_dir(file_name="media_downloads")
)


class _SharedAlbumOptionsInfo(TypedDict):
    isCollaborative: bool
    isCommentable: bool


class _ShareInfoInfo(TypedDict):
    isJoinable: bool
    isJoined: bool
    isOwned: bool
    shareableUrl: str
    sharedAlbumOptions: _SharedAlbumOptionsInfo
    shareToken: str


class _AlbumInfo(_GoogleEntityInfo):
    coverPhotoBaseUrl: str
    coverPhotoMediaItemId: str
    id: str
    isWriteable: bool
    mediaItemsCount: str
    productUrl: str
    shareInfo: _ShareInfoInfo
    title: str


class _MediaItemMetadataInfo(TypedDict):
    creationTime: str
    height: str
    width: str


class _MediaItemInfo(_GoogleEntityInfo):
    baseUrl: str
    contributorInfo: dict[str, str]
    description: str
    filename: str
    id: str
    mediaMetadata: _MediaItemMetadataInfo
    mimeType: str
    productUrl: str


class MediaType(Enum):
    """Enum for all potential media types"""

    IMAGE = auto()
    VIDEO = auto()
    UNKNOWN = auto()


class Album:
    """Class for Google Photos albums and their metadata/content

    Args:
        json (dict): the JSON which is returned from the Google Photos API when
         describing the album
        google_client (GooglePhotosClient): an active Google client which can be used to
         list media items

    """

    def __init__(self, json: _AlbumInfo, google_client: GooglePhotosClient):
        self.json = json
        self._media_items: list[MediaItem] | None = None

        self.google_client = google_client

    @property
    def media_items(self) -> list[MediaItem]:
        # noinspection GrazieInspection
        """Lists all media items in the album

        Returns:
            list: a list of MediaItem instances, representing the contents of the album
        """

        if not self._media_items:
            self._media_items = self.google_client.get_album_contents(self.id)

        return self._media_items

    @property
    def title(self) -> str | None:
        """
        Returns:
            str: the title of the album
        """
        return self.json.get("title")

    @property
    def id(self) -> str:
        """
        Returns:
            str: the ID of the album
        """
        return self.json["id"]

    @property
    def media_items_count(self) -> int:
        """
        Returns:
            int: the number of media items within the album
        """
        return int(self.json.get("mediaItemsCount", "-1"))

    def __str__(self) -> str:
        return f"{self.title}: {self.id}"


class MediaItem:
    """Class for representing a MediaItem and its metadata/content

    Args:
        json (dict): the JSON returned from the Google Photos API, which describes
        this media item
    """

    def __init__(self, json: _MediaItemInfo):
        self.json: _MediaItemInfo = json

        try:
            self.creation_datetime = datetime.strptime(
                json["mediaMetadata"]["creationTime"],
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )
        except ValueError:
            self.creation_datetime = datetime.strptime(
                json["mediaMetadata"]["creationTime"], "%Y-%m-%dT%H:%M:%SZ"
            )

        self.local_path = join(
            LOCAL_MEDIA_DIRECTORY,
            self.creation_datetime.strftime("%Y/%m/%d"),
            json["filename"],
        )

    def download(
        self,
        width_override: int | None = None,
        height_override: int | None = None,
        force_download: bool = False,
    ) -> str:
        """Download the media item to local storage. The width/height overrides do
        not apply to videos

        Args:
            width_override (int): the width override to use when downloading the file
            height_override (int): the height override to use when downloading the file
            force_download (bool): flag for forcing a download, even if it exists
             locally already

         Returns:
             str: the path to the downloaded file (self.local_path)
        """
        if not self.stored_locally or force_download:
            width = width_override or self.width
            height = height_override or self.height

            # LOGGER.debug("Downloading %s (%sx%s)", self.filename, width, height)

            param_str = {
                MediaType.IMAGE: f"=w{width}-h{height}",
                MediaType.VIDEO: "=dv",
            }.get(self.media_type, "")

            with open(force_mkdir(self.local_path, path_is_file=True), "wb") as fout:
                fout.write(get(f"{self.json['baseUrl']}{param_str}").content)

        return self.local_path

    @property
    def bytes(self) -> bytes:
        """Opens the local copy of the file (downloading it first if necessary) and
        reads the binary content of it

        Returns:
            bytes: the binary content of the file
        """
        if not self.stored_locally:
            self.download()

        with open(self.local_path, "rb") as fin:
            bytes_content = fin.read()

        return bytes_content

    @property
    def stored_locally(self) -> bool:
        """
        Returns:
            bool: flag for if the file exists locally
        """
        return isfile(self.local_path)

    @property
    def filename(self) -> str | None:
        """
        Returns:
            str: the media item's file name
        """
        return self.json.get("filename")

    @property
    def height(self) -> int:
        """
        Returns:
            int: the media item's height
        """
        return int(self.json.get("mediaMetadata", {}).get("height", "-1"))

    @property
    def width(self) -> int:
        """
        Returns:
            int: the media item's width
        """
        return int(self.json.get("mediaMetadata", {}).get("width", "-1"))

    @property
    def media_type(self) -> MediaType:
        """Determines the media item's file type from the JSON

        Returns:
            MediaType: the media type (image, video, etc.) for this item
        """

        if "image" in (mime_type := self.json.get("mimeType", "")):
            return MediaType.IMAGE

        if "video" in mime_type:
            return MediaType.VIDEO

        return MediaType.UNKNOWN

    def __str__(self) -> str:
        return dumps(self.json, indent=4, default=str)


class GooglePhotosClient(GoogleClient):
    """Custom client for interacting with the Google Photos API

    See Also:
        GoogleClient: the base Google client, used for authentication and common
         functions
    """

    BASE_URL = "https://photoslibrary.googleapis.com/v1"

    def __init__(
        self,
        project: str,
        scopes: list[str] | None = None,
        client_id_json_path: str | None = None,
        creds_cache_path: str | None = None,
        access_token_expiry_threshold: int = 60,
        logger: Logger | None = None,
    ):
        super().__init__(
            project,
            scopes,
            client_id_json_path,
            creds_cache_path,
            access_token_expiry_threshold,
            logger,
        )

        self._albums: list[Album] | None = None

    def get_album_contents(self, album_id: str) -> list[MediaItem]:
        # noinspection GrazieInspection
        """Gets the contents of a given album in Google Photos

        Args:
            album_id (str): the ID of the album which we want the contents of

        Returns:
            list: a list of MediaItem instances, representing each item in the album
        """

        return [
            MediaItem(item)
            for item in cast(
                Iterable[_MediaItemInfo],
                self._list_items(
                    self.session.post,
                    f"{self.BASE_URL}/mediaItems:search",
                    "mediaItems",
                    params={"albumId": album_id},
                ),
            )
        ]

    def get_album_from_name(self, album_name: str) -> Album:
        """Gets an album definition from the Google API based on the album name

        Args:
            album_name (str): the name of the album to find

        Returns:
            Album: an Album instance, with all metadata etc.

        Raises:
            FileNotFoundError: if the client can't find an album with the correct name
        """

        self.logger.info("Getting metadata for album `%s`", album_name)
        for album in self.albums:
            if album.title == album_name:
                return album

        raise FileNotFoundError(f"Unable to find album with name {album_name}")

    @property
    def albums(self) -> list[Album]:
        """Lists all albums in the active Google account

        Returns:
            list: a list of Album instances
        """

        if not self._albums:
            self._albums = [
                Album(item, self)
                for item in cast(
                    Iterable[_AlbumInfo],
                    self._list_items(
                        self.session.get,
                        f"{self.BASE_URL}/albums",
                        "albums",
                    ),
                )
            ]

        return self._albums



"""Custom client for interacting with Google's Drive API"""
from __future__ import annotations

from copy import deepcopy
from logging import Logger
from typing import Any, Collection, Dict, Iterable, cast

from wg_utilities.clients._google import GoogleClient
from wg_utilities.exceptions import ResourceNotFound


class File:
    """A file object within Google Drive

    Args:
        id (str): the unique ID of this file
        name (str): the name of the file
        parents (list): a list of parent directory IDs (should always be of length 1)
        google_client (GoogleDriveClient): a Google client for use in other requests

    Attributes:
        file_id (str): the unique ID of this file
        name (str): the name of the file
        parent_id (str): the unique ID of the parent directory
        google_client (GoogleDriveClient): a Google client for use in other requests

    """

    DESCRIPTION_FIELDS = [
        "appProperties",
        "capabilities",
        "contentHints",
        "contentRestrictions",
        "copyRequiresWriterPermission",
        "createdTime",
        "description",
        "driveId",
        "explicitlyTrashed",
        "exportLinks",
        "fileExtension",
        "folderColorRgb",
        "fullFileExtension",
        "hasAugmentedPermissions",
        "hasThumbnail",
        "headRevisionId",
        "iconLink",
        "id",
        "imageMediaMetadata",
        "isAppAuthorized",
        "kind",
        "lastModifyingUser",
        "linkShareMetadata",
        "md5Checksum",
        "mimeType",
        "modifiedByMe",
        "modifiedByMeTime",
        "modifiedTime",
        "name",
        "originalFilename",
        "ownedByMe",
        "owners",
        "parents",
        "permissionIds",
        "permissions",
        "properties",
        "quotaBytesUsed",
        "resourceKey",
        "shared",
        "sharedWithMeTime",
        "sharingUser",
        "shortcutDetails",
        "size",
        "spaces",
        "starred",
        "teamDriveId",
        "thumbnailLink",
        "thumbnailVersion",
        "trashed",
        "trashedTime",
        "trashingUser",
        "version",
        "videoMediaMetadata",
        "viewedByMe",
        "viewedByMeTime",
        "viewersCanCopyContent",
        "webContentLink",
        "webViewLink",
        "writersCanShare",
    ]

    # noinspection PyShadowingBuiltins
    # pylint: disable=redefined-builtin
    def __init__(
        self,
        *,
        id: str,
        name: str,
        google_client: GoogleDriveClient,
        parents: list[str] | None = None,
        **_: Any,
    ):
        self.file_id = id
        self.name = name
        self.parent_id = parents.pop() if parents else None
        self.google_client = google_client

        self._description: dict[str, str | bool | float | int]
        self._parent: Directory

    def describe(
        self, force_update: bool = False
    ) -> dict[str, str | bool | float | int]:
        """Describe the file by requesting all available fields from the Drive API

        Args:
            force_update (bool): re-pull the description from Google Drive, even if we
             already have the description locally

        Returns:
            dict: the description JSON for this file
        """
        if force_update or not hasattr(self, "_description"):
            self._description = self.google_client.session.get(
                f"{self.google_client.BASE_URL}/files/{self.file_id}",
                params={"fields": ", ".join(self.DESCRIPTION_FIELDS)},
            ).json()

        return self._description

    @property
    def parent(self) -> Directory:
        """
        Returns:
            Directory: the parent directory of this file
        """
        if not hasattr(self, "_parent"):
            self._parent = self.google_client.get_directory_by(
                "file_id", self.parent_id
            )

        return self._parent

    @property
    def path(self) -> str:
        """
        Returns:
            str: the path to this file in Google Drive
        """
        current_path = self.name

        parent_dir = self.parent
        while True:
            current_path = "/".join([parent_dir.name, current_path])

            if not (next_parent := parent_dir.parent):
                break

            parent_dir = next_parent

        return "/" + current_path

    def __gt__(self, other: File) -> bool:
        return self.name.lower() > other.name.lower()

    def __lt__(self, other: File) -> bool:
        return self.name.lower() < other.name.lower()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name


class Directory(File):
    """A Google Drive directory - basically a File with extended functionality

    Attributes:
        children (set): the directories contained within this directory
    """

    # noinspection PyShadowingBuiltins
    # pylint: disable=redefined-builtin
    def __init__(
        self,
        *,
        id: str,
        name: str,
        google_client: GoogleDriveClient,
        parents: list[str] | None = None,
        **_: Any,
    ):
        super().__init__(id=id, name=name, parents=parents, google_client=google_client)
        self.children: set[Directory] = set()
        self._files: list[File] | None = None

    @property
    def files(self) -> list[File]:
        """
        Returns:
            list: the list of files contained within this directory
        """
        if not self._files:
            self._files = [
                File(**item, google_client=self.google_client)  # type: ignore[arg-type]
                # TODO make this just pass in the JSON like everything else...
                for item in cast(
                    Iterable[Dict[str, str]],
                    self.google_client.get_items(
                        f"{self.google_client.BASE_URL}/files",
                        "files",
                        params={
                            "pageSize": "1000",
                            "q": "mimeType != 'application/vnd.google-apps.folder' and"
                            f" '{self.file_id}' in parents",
                            "fields": "nextPageToken, files(id, name, parents)",
                        },
                    ),
                )
            ]

        return list(self._files)

    @property
    def parent(self) -> Directory:
        """
        Returns:
            Directory: the parent directory of this directory
        """
        return self._parent

    @parent.setter
    def parent(self, value: Directory) -> None:
        self._parent = value

    @property
    def tree(self) -> str:
        """A simple copy of the Linux `tree` command, this builds a directory tree in
        text form for quick visualisation

        Returns:
            str: the full directory tree in text form
        """

        output = self.name

        def build_sub_tree(
            parent_dir: Directory,
            level: int,
            block_pipes_at_levels: list[int] | None = None,
            show_files: bool = True,
        ) -> None:
            """Builds a subtree of a given directory

            Args:
                parent_dir (Directory): the directory to create the subtree of
                level (int): the depth level of this directory
                block_pipes_at_levels (list): a list of levels to block further pipes at
                show_files (bool): flag to load files too - this could take
                 considerably longer to load (proportional to the number of files you've
                 got)
            """

            nonlocal output

            # Creating a deep copy means that when we go back up from the recursion,
            # the previous iteration still has the correct levels in the list
            block_pipes_at_levels = deepcopy(block_pipes_at_levels) or []

            item_list: Collection[File] = parent_dir.children

            if not item_list and show_files:
                item_list = parent_dir.files

            for i, child_item in enumerate(sorted(item_list)):
                prefix = "\n"

                # build out the spaces and pipes on this line, in such a way to
                # maintain continuity from the previous line
                for j in range(level):
                    prefix += " " if j in block_pipes_at_levels else "│"
                    prefix += "    "

                # if this is the last child
                if i + 1 == len(item_list):
                    prefix += "└"
                    block_pipes_at_levels.append(level)
                else:
                    prefix += "├"

                if isinstance(child_item, Directory):
                    prefix += "─── "
                else:
                    prefix += "--> "

                output += prefix + child_item.name

                if isinstance(child_item, Directory):
                    build_sub_tree(child_item, level + 1, block_pipes_at_levels)

        build_sub_tree(self, 0)

        return output


class GoogleDriveClient(GoogleClient):
    """Custom client specifically for Google's Drive API

    Args:
        project (str): the name of the project which this client is being used for
        scopes (list): a list of scopes the client can be given
        client_id_json_path (str): the path to the `client_id.json` file downloaded
         from Google's API Console
        creds_cache_path (str): file path for where to cache credentials
        access_token_expiry_threshold (int): the threshold for when the access token is
         considered expired
        logger (RootLogger): a logger to use throughout the client functions

    """

    BASE_URL = "https://www.googleapis.com/drive/v3"

    def __init__(
        self,
        project: str,
        scopes: list[str] | None = None,
        client_id_json_path: str | None = None,
        creds_cache_path: str | None = None,
        access_token_expiry_threshold: int = 60,
        logger: Logger | None = None,
    ):
        super().__init__(
            project=project,
            scopes=scopes,
            client_id_json_path=client_id_json_path,
            creds_cache_path=creds_cache_path,
            access_token_expiry_threshold=access_token_expiry_threshold,
            logger=logger,
        )
        self._root_directory: Directory

        self._directories: list[Directory]

    def _build_directory_structure(self) -> None:
        """Build the complete tree of directories, including parent-child relationships
        by listing all directories and then iterating through them to build the
        relationships
        """
        self._directories = [self.root_directory]

        # List every single directory
        directory: dict[str, str]
        for directory in self.get_items(
            f"{self.BASE_URL}/files",
            "files",
            params=dict(
                pageSize="1000",
                q="mimeType = 'application/vnd.google-apps.folder'",
                fields="nextPageToken, files(id, name, parents)",
            ),
        ):
            self._directories.append(Directory(**directory, google_client=self))

        # Iterate through the directories, adding their parents and children as
        # applicable
        for directory_ in self._directories:
            if parent_dir := self.get_directory_by("file_id", directory_.parent_id):
                directory_.parent = parent_dir
                parent_dir.children.add(directory_)

    def get_directory_by(self, attribute: str, value: Any) -> Directory:
        """Get a Directory instance by any attribute

        Args:
            attribute (str): the name of the attribute to search for
            value (Any): the target value of the attribute

        Returns:
            Directory: the directory being searched for, if it was found

        Raises:
            ResourceNotFound: if no matching directory exists
        """
        for directory in self.directories:
            if getattr(directory, attribute) == value:
                return directory

        raise ResourceNotFound(
            f"Unable to find directory where attribute {attribute!r} == {str(value)}"
        )

    def get_file_from_id(
        self,
        file_id: str,
        params: dict[str, str | int | float | bool] | None = None,
    ) -> File:
        """Find a file by its UUID

        Args:
            file_id (str): the unique ID of the file
            params (dict): any params to pass in the request

        Returns:
            File: the target file, if found
        """

        if not self._directories:
            self._build_directory_structure()

        params = params or {}

        if "fields" not in params:
            params["fields"] = "id, name, parents"

        return File(
            **self.session.get(
                f"{self.BASE_URL}/files/{file_id}", params=params
            ).json(),
            google_client=self,
        )

    @property
    def shared_drives(self) -> list[dict[str, str]]:
        """
        Returns:
            list: a list of Shared Drives the current user has access to
        """
        # pylint: disable=line-too-long
        return self.get_items(f"{self.BASE_URL}/drives", "drives")  # type: ignore[return-value]

    @property
    def directories(self) -> list[Directory]:
        """
        Returns:
            list: a list of all directories in the user's Drive
        """
        if not self._directories:
            self._build_directory_structure()

        return self._directories

    @property
    def root_directory(self) -> Directory:
        """
        Returns:
            Directory: the user's root directory/main Drive
        """
        if not hasattr(self, "_root_directory"):
            res = self.session.get(f"{self.BASE_URL}/files/root")
            res.raise_for_status()
            self._root_directory = Directory(
                **res.json(),
                google_client=self,
            )
            if not self._directories:
                self._build_directory_structure()

        return self._root_directory



"""Useful functions for working with JSON/dictionaries"""
from logging import DEBUG, getLogger
from typing import Any, Callable, Dict, List, Optional, Sequence, Type

LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)


def set_nested_value(
    *,
    json_obj: Dict[Any, Any],
    keys: List[str],
    target_value: Any,
    final_key: Optional[str] = None,
) -> None:
    """Update a nested value in a dictionary

    Args:
        json_obj (dict): the JSON object to update
        keys (list): a list of keys used to traverse the dictionary
        target_value (Any): the value to set at the given location/path
        final_key (str): the final key, the value of which we're actually setting
    """

    final_key = final_key or keys.pop()

    if len(keys) > 0:
        set_nested_value(
            json_obj=json_obj.get(keys.pop(0), {}),
            keys=keys,
            target_value=target_value,
            final_key=final_key,
        )
    else:
        json_obj[final_key] = target_value


def process_list(
    lst: List[Any],
    *,
    target_type: Type[Any],
    target_processor_func: Callable[[Any], Any],
    pass_on_fail: bool = True,
    log_op_func_failures: bool = False,
    single_keys_to_remove: Optional[Sequence[str]] = None,
) -> None:
    """Iterates through a list and applies `list_op_func` to any instances of
    `target_type`

    Args:
        lst (list): the list to iterate through
        target_type (type): the target type to apply functions to
        target_processor_func (Callable): a function to apply to instances of
         `target_type`
        pass_on_fail (bool): ignore failure in either op function
        log_op_func_failures (bool): log any failures in either op function
        single_keys_to_remove (list): a list of keys that can be "expanded" up to the
         parent key

    Raises:
        Exception: if the `target_processor_func` fails and `pass_on_fail` is False
    """
    for i, elem in enumerate(lst):
        if isinstance(elem, target_type):
            try:
                lst[i] = target_processor_func(elem)
            except Exception as exc:  # pylint: disable=broad-except
                if log_op_func_failures:
                    LOGGER.debug(str(exc))
                if not pass_on_fail:
                    raise
        elif isinstance(elem, dict):
            traverse_dict(
                elem,
                target_type=target_type,
                target_processor_func=target_processor_func,
                pass_on_fail=pass_on_fail,
                log_op_func_failures=log_op_func_failures,
                single_keys_to_remove=single_keys_to_remove,
            )
        elif isinstance(elem, list):
            process_list(
                elem,
                target_type=target_type,
                target_processor_func=target_processor_func,
                pass_on_fail=pass_on_fail,
                log_op_func_failures=log_op_func_failures,
            )


def traverse_dict(
    payload_json: Dict[Any, Any],
    *,
    target_type: Type[Any],
    target_processor_func: Callable[[Any], Any],
    pass_on_fail: bool = True,
    log_op_func_failures: bool = False,
    single_keys_to_remove: Optional[Sequence[str]] = None,
) -> None:
    # pylint: disable=too-many-branches,too-many-nested-blocks
    """Recursively traverse a JSON object and apply `dict_op_func` to any values
    of type `target_type`

    Args:
        payload_json (dict): the JSON object to traverse
        target_type (type): the target type to apply functions to
        target_processor_func (Callable): a function to apply to instances of
         `target_type`
        pass_on_fail (bool): ignore failure in either op function
        log_op_func_failures (bool): log any failures in either op function
        single_keys_to_remove (list): a list of keys that can be "expanded" up to the
         parent key, e.g.:
            ... {
            ...     "parent_1": "something",
            ...     "parent_2": {
            ...         "val": "actual value"
            ...     }
            ... }
            would go to
            ... {
            ...     "parent_1": "something",
            ...     "parent_2": "actual value"
            ... }

    Raises:
        Exception: if the `target_processor_func` fails and `pass_on_fail` is False
    """
    for k, v in payload_json.items():
        if isinstance(v, target_type):
            try:
                payload_json.update({k: target_processor_func(v)})
                if isinstance(payload_json.get(k), dict):
                    traverse_dict(
                        payload_json,
                        target_type=target_type,
                        target_processor_func=target_processor_func,
                        pass_on_fail=pass_on_fail,
                        log_op_func_failures=log_op_func_failures,
                        single_keys_to_remove=single_keys_to_remove,
                    )
            except Exception as exc:  # pylint: disable=broad-except
                if log_op_func_failures:
                    LOGGER.error(str(exc))
                if not pass_on_fail:
                    raise
        elif isinstance(v, dict):
            matched_single_key = False
            if len(v) == 1 and single_keys_to_remove is not None:
                if (only_key := next(iter(v.keys()))) in single_keys_to_remove:
                    matched_single_key = True
                    if isinstance(value := v.get(only_key), target_type):
                        try:
                            value = target_processor_func(value)

                            if isinstance(value, dict):
                                # Wrap the value, so that if the top level key is one
                                # of `single_keys_to_remove` then it's processed
                                # correctly
                                tmp_wrapper = {"-": value}
                                traverse_dict(
                                    tmp_wrapper,
                                    target_type=target_type,
                                    target_processor_func=target_processor_func,
                                    pass_on_fail=pass_on_fail,
                                    log_op_func_failures=log_op_func_failures,
                                    single_keys_to_remove=single_keys_to_remove,
                                )

                                value = tmp_wrapper["-"]
                        except Exception as exc:  # pylint: disable=broad-except
                            if log_op_func_failures:
                                LOGGER.error(str(exc))
                            if not pass_on_fail:
                                raise
                    payload_json[k] = value

            if not matched_single_key:
                traverse_dict(
                    v,
                    target_type=target_type,
                    target_processor_func=target_processor_func,
                    pass_on_fail=pass_on_fail,
                    log_op_func_failures=log_op_func_failures,
                    single_keys_to_remove=single_keys_to_remove,
                )
        elif isinstance(v, list):
            process_list(
                v,
                target_type=target_type,
                target_processor_func=target_processor_func,
                pass_on_fail=pass_on_fail,
                log_op_func_failures=log_op_func_failures,
                single_keys_to_remove=single_keys_to_remove,
            )



"""Generic Google Client - having one client for all APIs is way too big"""
from __future__ import annotations

from copy import deepcopy
from json import dump, dumps, load
from logging import DEBUG, Logger, getLogger
from os import remove
from time import time
from typing import Any, Literal, Mapping, TypedDict, TypeVar
from webbrowser import open as open_browser

from google.auth.transport.requests import AuthorizedSession
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import Flow
from jwt import DecodeError, decode
from requests import Response, post
from typing_extensions import Protocol

from wg_utilities.clients._generic import TempAuthServer
from wg_utilities.functions import force_mkdir, user_data_dir


class _GoogleEntityInfo(TypedDict):
    pass


_GoogleEntityInfoTypeVar = TypeVar("_GoogleEntityInfoTypeVar", bound=_GoogleEntityInfo)


class _GoogleCredentialsInfo(TypedDict):
    access_token: str
    client_id: str
    client_secret: str
    expires_in: int
    id_token: str | None
    refresh_token: str
    scope: str
    scopes: list[str]
    token: str
    token_type: Literal["Bearer"]
    token_uri: str


class _SessionMethodCallable(Protocol):
    def __call__(self, url: str, params: dict[str, Any] | None = None) -> Response:
        ...


class GoogleClient:
    """Custom client for interacting with the Google APIs

    Args:
        project (str): the name of the project which this client is being used for
        scopes (list): a list of scopes the client can be given
        client_id_json_path (str): the path to the `client_id.json` file downloaded
         from Google's API Console
        creds_cache_path (str): file path for where to cache credentials
        access_token_expiry_threshold (int): the threshold for when the access token is
         considered expired
        logger (RootLogger): a logger to use throughout the client functions
    """

    DEFAULT_PARAMS = {
        "pageSize": "50",
    }

    def __init__(
        self,
        project: str,
        scopes: list[str] | None = None,
        client_id_json_path: str | None = None,
        creds_cache_path: str | None = None,
        access_token_expiry_threshold: int = 60,
        logger: Logger | None = None,
    ):
        self.project = project
        self.scopes = scopes or []
        self.client_id_json_path = client_id_json_path
        self.creds_cache_path = creds_cache_path or user_data_dir(
            file_name="google_api_creds.json"
        )
        self.access_token_expiry_threshold = access_token_expiry_threshold

        if logger:
            self.logger = logger
        else:
            self.logger = getLogger(__name__)
            self.logger.setLevel(DEBUG)

        if not scopes:
            self.logger.warning(
                "No scopes set for Google client. Functionality will be limited."
            )

        self._all_credentials_json: dict[str, _GoogleCredentialsInfo] = {}
        self._session = None

        self.temp_auth_server: TempAuthServer | None = None

    def _list_items(
        self,
        method: _SessionMethodCallable,
        url: str,
        list_key: str,
        *,
        params: Mapping[str, Any] | None = None,
    ) -> list[_GoogleEntityInfoTypeVar]:
        """Generic method for listing items on Google's API(s)

        Args:
            method (Callable): the Google client session method to use
            url (str): the API endpoint to send a request to
            list_key (str): the key to use in extracting the data from the response
            params (dict): any extra params to be passed in the request

        Returns:
            list: a list of dicts, each representing an item from the API
        """
        params = (
            {**self.DEFAULT_PARAMS, **params}
            if params
            else deepcopy(self.DEFAULT_PARAMS)
        )
        self.logger.info(
            "Listing all items at endpoint `%s` with params %s", url, dumps(params)
        )

        res = method(url, params=params)

        res.raise_for_status()

        item_list: list[_GoogleEntityInfoTypeVar] = res.json().get(list_key, [])

        while next_token := res.json().get("nextPageToken"):
            # noinspection PyArgumentList
            res = method(
                url,
                params={**params, "pageToken": next_token},
            )
            res.raise_for_status()
            item_list.extend(res.json().get(list_key, []))
            self.logger.debug("Found %i items so far", len(item_list))

        return item_list

    def delete_creds_file(self) -> None:
        """Delete the local creds file"""
        try:
            remove(self.creds_cache_path)
        except FileNotFoundError:
            pass

    def get_items(
        self,
        url: str,
        list_key: str = "items",
        *,
        params: Mapping[str, str | int | float | bool] | None = None,
    ) -> list[_GoogleEntityInfoTypeVar]:
        """Wrapper method for getting a list of items

        See Also:
            self._list_items: main worker method for this functionality
        """

        return self._list_items(self.session.get, url, list_key, params=params)

    def refresh_access_token(self) -> None:
        """Uses the cached refresh token to submit a request to TL's API for a new
        access token"""

        self.logger.info("Refreshing access token")

        res = post(
            "https://accounts.google.com/o/oauth2/token",
            data={
                "grant_type": "refresh_token",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "refresh_token": self.refresh_token,
            },
        )

        res.raise_for_status()

        # Maintain any existing credential values in the dictionary whilst updating
        # new ones
        self.credentials = {
            **self._all_credentials_json[self.project],
            **res.json(),
        }

    @property
    def access_token_has_expired(self) -> bool:
        """Decodes the JWT access token and evaluates the expiry time

        Returns:
            bool: has the access token expired?
        """

        if self.project not in self._all_credentials_json:
            return True

        try:
            expiry_epoch = decode(
                # can't use self.access_token here, as that uses self.credentials,
                # which in turn (recursively) checks if the access token has expired
                self._all_credentials_json[self.project]["access_token"],
                options={"verify_signature": False},
            ).get("exp", 0)

            return bool(
                (expiry_epoch - self.access_token_expiry_threshold) < int(time())
            )
        except (DecodeError, KeyError):
            # treat invalid/missing token as expired, so we get a new one
            return True

    @property
    def client_id(self) -> str:
        """
        Returns:
            str: the current client ID
        """
        return self._all_credentials_json[self.project]["client_id"]

    @property
    def client_secret(self) -> str:
        """
        Returns:
            str: the current client secret
        """
        return self._all_credentials_json[self.project]["client_secret"]

    @property
    def credentials(self) -> Credentials:
        """Gets creds as necessary (including first time setup) and authenticates them

        Returns:
            Credentials: authorized credentials for use in creating a session

        Raises:
            EOFError: when no data is successfully returned for the auth code (usually
             when running the script automatically)
            ValueError: same as above, but if the EOFError isn't raised
        """

        try:
            with open(
                force_mkdir(self.creds_cache_path, path_is_file=True), encoding="UTF-8"
            ) as fin:
                self._all_credentials_json = load(fin)
        except FileNotFoundError:
            self.logger.info("Unable to find local creds file")
            self._all_credentials_json = {}

        if self.project not in self._all_credentials_json:
            self.logger.info(
                "Performing first time login for project `%s`", self.project
            )

            self.client_id_json_path = self.client_id_json_path or input(
                "Download your Client ID JSON from https://console.cloud.google.com/"
                f"apis/credentials?project={self.project} and paste the file path"
                " here: "
            )

            flow = Flow.from_client_secrets_file(
                self.client_id_json_path,
                scopes=self.scopes,
                redirect_uri="http://localhost:5001/get_auth_code",
            )

            auth_url, _ = flow.authorization_url(access_type="offline")
            self.logger.debug("Opening %s", auth_url)
            open_browser(auth_url)

            self.temp_auth_server = TempAuthServer(__name__)

            request_args = self.temp_auth_server.wait_for_request(
                "/get_auth_code", kill_on_request=True
            )

            flow.fetch_token(code=request_args.get("code"))

            self.credentials = {
                "token": flow.credentials.token,
                "refresh_token": flow.credentials.refresh_token,
                "id_token": flow.credentials.id_token,
                "scopes": flow.credentials.scopes,
                "token_uri": flow.credentials.token_uri,
                "client_id": flow.credentials.client_id,
                "client_secret": flow.credentials.client_secret,
            }

        if self.access_token_has_expired:
            self.refresh_access_token()

        return Credentials.from_authorized_user_info(
            self._all_credentials_json[self.project], self.scopes
        )

    @credentials.setter
    def credentials(self, value: _GoogleCredentialsInfo) -> None:
        """
        Args:
            value (dict): the new values to use for the creds for this project
        """
        self._all_credentials_json[self.project] = value

        with open(
            force_mkdir(self.creds_cache_path, path_is_file=True), "w", encoding="UTF-8"
        ) as fout:
            dump(self._all_credentials_json, fout)

    @property
    def refresh_token(self) -> str:
        """
        Returns:
            str: the current refresh token
        """
        return self._all_credentials_json[self.project]["refresh_token"]

    @property
    def session(self) -> AuthorizedSession:
        """Uses the Credentials object to sign in to an authorized Google API session

        Returns:
            AuthorizedSession: an active, authorized Google API session
        """
        if not self._session:
            self._session = AuthorizedSession(self.credentials)

        return self._session



"""This module contains any custom mocks (classes or functions) for use in Unit Tests
within this project and others"""
from typing import Any, Callable, Dict, List, Optional, Union

from botocore.client import BaseClient

# noinspection PyProtectedMember
# pylint: disable=protected-access
ORIG_API_CALL: Callable[..., Dict[str, str]] = BaseClient._make_api_call  # type: ignore


class MockBoto3Client:
    """Class for adding custom mocks for boto3 when moto doesn't support the
    operation"""

    def __init__(
        self,
        mocked_operation_lookup: Optional[
            Dict[str, Union[str, Callable[..., Any]]]
        ] = None,
    ):
        self.mocked_operation_lookup = mocked_operation_lookup or {}

        self.boto3_calls: Dict[str, List[Dict[Any, Any]]] = {}

    def reset_boto3_calls(self) -> None:
        """Resets the boto3 calls to an empty dict"""
        self.boto3_calls = {}

    def build_api_call(
        self,
        lookup_overrides: Optional[Dict[str, Union[str, Callable[..., Any]]]] = None,
        reset_boto3_calls: bool = True,
    ) -> Callable[[BaseClient, str, Dict[str, Any]], Union[Dict[str, str], str]]:
        """Wrapper function for the API call. Also resets the internal log of boto3
        calls as this is a new API call

        Args:
            lookup_overrides (dict): any overrides to be applied for this specific API
             call
            reset_boto3_calls (bool): option for resetting boto3 calls

        Returns:
            function: the mocked API call
        """
        lookup_overrides_ = lookup_overrides or {}

        if reset_boto3_calls:
            self.reset_boto3_calls()

        def api_call(
            client: BaseClient, operation_name: str, kwargs: Dict[str, Any]
        ) -> Union[Dict[str, str], str]:
            """Inner function of this mock, which is the actual mock function itself.

            Args:
                client (BaseClient): the client making the (mocked) request
                operation_name (str): the AWS operation being requested
                kwargs (dict): any keyword arguments being passed to AWS

            Returns:
                dict: a (mocked) response from AWS

            Raises:
                Exception: if an operation is requested that isn't mocked by this
                 class or by moto
                KeyError: when a KeyError is caught, but it isn't for the above
                 reason, it's just re-raised
            """

            # If the response override is a function (so we can dynamically set the
            # response value) then call it - otherwise just return it
            # This needs to be in here because if the override _is_ callable,
            # then we need to call it on each API call to update values etc.
            for operation, response_override in lookup_overrides_.items():
                self.mocked_operation_lookup[operation] = (
                    response_override()
                    if callable(response_override)
                    else response_override
                )

            self.boto3_calls.setdefault(operation_name, []).append(kwargs)

            if (
                mocked_operation := self.mocked_operation_lookup.get(operation_name)
            ) is not None:
                if callable(mocked_operation):
                    return mocked_operation(kwargs)  # type: ignore

                return mocked_operation

            try:
                return ORIG_API_CALL(client, operation_name, kwargs)
            except KeyError as exc:
                if str(exc) == "'DEFAULT'":
                    raise Exception(
                        f"Operation `{operation_name}` not supported by moto yet"
                    ) from exc
                raise

        return api_call



"""Useful functions"""

from ._functions import chunk_list, flatten_dict, try_float
from .datetime_helpers import utcnow
from .file_management import force_mkdir, user_data_dir
from .json import process_list, set_nested_value, traverse_dict
from .processes import run_cmd
from .string_manipulation import cleanse_string

__all__ = [
    "chunk_list",
    "flatten_dict",
    "try_float",
    "utcnow",
    "force_mkdir",
    "user_data_dir",
    "set_nested_value",
    "run_cmd",
    "cleanse_string",
    "process_list",
    "traverse_dict",
]

try:
    from .xml import get_nsmap

    __all__.append(
        "get_nsmap",
    )
except ImportError as _exc:
    if str(_exc) == "No module named 'lxml'":
        pass



"""Custom exception types"""
from __future__ import annotations

from functools import wraps
from inspect import stack
from logging import Logger
from os import getenv
from socket import gethostname
from traceback import format_exc
from typing import Any, Callable

from dotenv import load_dotenv
from requests import post

load_dotenv()

HA_LOG_ENDPOINT = getenv("HA_LOG_ENDPOINT", "192.168.1.120:8001")


class ResourceNotFound(Exception):
    """Custom exception for when some kind of resource isn't found"""


def send_exception_to_home_assistant(exc: Exception) -> None:
    """Format an exception and send useful info to Home Assistant

    Args:
        exc (Exception): the exception being handled

    Raises:
        ValueError: if the HA_LOG_ENDPOINT isn't set
        Exception: if posting the exception to HA fails, then an exception is raised
    """
    payload = {
        "client": gethostname(),
        "message": f"{type(exc).__name__} in `{stack()[2].filename}`: {repr(exc)}",
        "traceback": format_exc(),
    }

    try:
        try:
            # If the host is on the same network as HA, then an HTTP local URL can be
            # used
            post(f"http://{HA_LOG_ENDPOINT}/log/error", json=payload)
        except Exception:  # pylint: disable=broad-except
            post(f"https://{HA_LOG_ENDPOINT}/log/error", json=payload)
    except Exception as send_exc:
        raise send_exc from exc


def on_exception(
    exception_callback: Callable[[Exception], Any] = send_exception_to_home_assistant,
    *,
    raise_after_callback: bool = True,
    logger: Logger | None = None,
) -> Callable[[Any], Any]:
    """Decorator factory to allow parameterizing the inner decorator

    Args:
        exception_callback (Callable): callback function to process the exception
        raise_after_callback (bool): raise the exception after the callback has run
        logger (Logger): optional logger for logging the exception

    Returns:
        Callable: the actual decorator
    """

    def _decorator(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        """Decorator to allow simple cover-all exception handler callback behaviour

        Args:
            func (Callable): the function being wrapped

        Returns:
            Callable: the inner function
        """

        @wraps(func)
        def worker(*args: Any, **kwargs: Any) -> Any:
            """Tries to run the decorated function and calls the callback function

            Args:
                *args (Any): any args passed to the inner func
                **kwargs (Any): any kwargs passed to the inner func

            Returns:
                Any: the result of the wrapped function

            Raises:
                Exception: any exception from the decorated function
            """

            try:
                return func(*args, **kwargs)
            except Exception as exc:  # pylint: disable=broad-except
                if logger is not None:
                    logger.exception(repr(exc))

                exception_callback(exc)

                if raise_after_callback:
                    raise

                return None

        return worker

    return _decorator


__all__ = ["on_exception", "send_exception_to_home_assistant", "ResourceNotFound"]



"""Custom client for interacting with Google's Fit API"""
from __future__ import annotations

from datetime import datetime
from logging import Logger
from typing import Literal, TypedDict

from wg_utilities.clients._google import GoogleClient, _GoogleEntityInfo
from wg_utilities.functions.datetime_helpers import DatetimeFixedUnit as DFUnit
from wg_utilities.functions.datetime_helpers import utcnow


class _DataSourceDataTypeFieldInfo(TypedDict):
    format: Literal[
        "blob", "floatList", "floatPoint", "integer", "integerList", "map", "string"
    ]
    name: str


class _DataSourceDataTypeInfo(TypedDict):
    field: list[_DataSourceDataTypeFieldInfo]
    name: str


class _DataSourceDescriptionInfo(TypedDict):
    application: dict[str, str]
    dataStreamId: str
    dataStreamName: str
    dataType: _DataSourceDataTypeInfo


class _GoogleFitDataPointInfo(_GoogleEntityInfo):
    id: str
    startTimeNanos: str
    endTimeNanos: str
    dataTypeName: str
    originDataSourceId: str
    value: list[dict[str, int]]


class DataSource:
    """Class for interacting with Google Fit Data Sources. An example of a data source
    is Strava, Google Fit, MyFitnessPal, etc. The ID is something _like_ "...weight",
    "...calories burnt"

    Args:
        data_source_id (str): the unique ID of the data source
        google_client (GoogleClient): a GoogleClient instance, needed for getting
         DataSource info
    """

    DP_VALUE_KEY_LOOKUP = {"floatPoint": "fpVal", "integer": "intVal"}

    def __init__(self, data_source_id: str, *, google_client: GoogleClient):
        self.data_source_id = data_source_id
        self.url = (
            f"{GoogleFitClient.BASE_URL}/users/me/dataSources/{self.data_source_id}"
        )
        self.google_client = google_client

        self._description: _DataSourceDescriptionInfo

    @property
    def description(self) -> _DataSourceDescriptionInfo:
        """
        Returns:
            dict: the JSON description of this data source
        """
        if not hasattr(self, "_description"):
            self._description = self.google_client.session.get(self.url).json()

        return self._description

    def sum_data_points_in_range(
        self,
        from_datetime: datetime | None = None,
        to_datetime: datetime | None = None,
    ) -> int:
        """Gets the sum of data points in the given range: if no `from_datetime` is
        provided, it defaults to the start of today; if no `to_datetime` is provided
        then it defaults to now.

        Args:
            from_datetime (datetime): lower boundary for step count
            to_datetime (datetime): upper boundary for step count

        Returns:
            int: a sum of data points in the given range
        """

        from_nano = int(
            int(from_datetime.timestamp() * 1000000000)
            if from_datetime
            else int(
                datetime.today()
                .replace(hour=0, minute=0, second=0, microsecond=0)
                .timestamp()
                / DFUnit.NANOSECOND.value
            )
        )

        to_nano = int(
            int(to_datetime.timestamp() * 1000000000)
            if to_datetime
            else utcnow(DFUnit.NANOSECOND)  # type: ignore[arg-type]
        )

        data_points: list[_GoogleFitDataPointInfo] = self.google_client.get_items(
            f"{self.url}/datasets/{from_nano}-{to_nano}",
            "point",
        )

        count = 0
        for point in data_points:
            if (
                int(point["startTimeNanos"]) >= from_nano
                and int(point["endTimeNanos"]) <= to_nano
            ):
                count += point["value"][0][self.data_point_value_key]

        return count

    @property
    def data_type_field_format(
        self,
    ) -> Literal[
        "blob", "floatList", "floatPoint", "integer", "integerList", "map", "string"
    ]:
        """
        Returns:
            str: the field format of this data source (i.e. "integer" or "floatPoint")

        Raises:
            Exception: if more than 1 dataType field value is found
        """
        data_type_fields = self.description["dataType"]["field"]
        if len(data_type_fields) != 1:
            raise Exception(
                f"Unexpected number of dataType fields ({len(data_type_fields)}): "
                + ", ".join(f["name"] for f in data_type_fields)
            )

        return data_type_fields[0]["format"]

    @property
    def data_point_value_key(self) -> Literal["fpVal", "intVal"]:
        """
        Returns:
            str: the key to use when extracting data from a data point
        """
        # pylint: disable=line-too-long
        return self.DP_VALUE_KEY_LOOKUP[self.data_type_field_format]  # type: ignore[return-value]


class GoogleFitClient(GoogleClient):
    """Custom client for interacting with the Google Fit API

    See Also:
        GoogleClient: the base Google client, used for authentication and common
         functions
    """

    BASE_URL = "https://www.googleapis.com/fitness/v1"

    def __init__(
        self,
        project: str,
        scopes: list[str] | None = None,
        client_id_json_path: str | None = None,
        creds_cache_path: str | None = None,
        access_token_expiry_threshold: int = 60,
        logger: Logger | None = None,
    ):
        super().__init__(
            project,
            scopes=scopes,
            client_id_json_path=client_id_json_path,
            creds_cache_path=creds_cache_path,
            access_token_expiry_threshold=access_token_expiry_threshold,
            logger=logger,
        )
        self.data_sources: dict[str, DataSource] = {}

    def get_data_source(self, data_source_id: str) -> DataSource:
        """Gets a data source based on its UID. DataSource instances are cached for the
         lifetime of the GoogleClient instance

        Args:
            data_source_id (str): the UID of the data source

        Returns:
            DataSource: an instance, ready to use!
        """

        # TODO why isn't this just a list an we compare the istance IDs?
        if (data_source := self.data_sources.get(data_source_id)) is None:
            data_source = DataSource(data_source_id, google_client=self)
            self.data_sources[data_source_id] = data_source

        return data_source



"""Custom client for interacting with Monzo's API"""
from __future__ import annotations

from datetime import datetime, timedelta
from logging import DEBUG, getLogger
from typing import Generator, Literal, TypedDict

from requests import get, put

from wg_utilities.clients._generic import OauthClient
from wg_utilities.functions import cleanse_string, user_data_dir

LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class _MonzoAccountInfo(TypedDict):
    account_number: str
    balance: float
    balance_including_flexible_savings: float
    created: str
    description: str
    id: str
    sort_code: str
    spend_today: float
    total_balance: float


class _MonzoPotInfo(TypedDict):
    available_for_bills: bool
    balance: float
    charity_id: str
    cover_image_url: str
    created: str
    currency: str
    current_account_id: str
    deleted: bool
    goal_amount: float
    has_virtual_cards: bool
    id: str
    is_tax_pot: bool
    isa_wrapper: str
    locked: bool
    name: str
    product_id: str
    round_up: bool
    round_up_multiplier: float
    style: str
    type: str
    updated: str


class _BalanceVariablesInfo(TypedDict):
    balance: int
    balance_including_flexible_savings: int
    currency: Literal["GBP"]
    local_currency: str
    local_exchange_rate: int | float
    local_spend: dict[str, str | int]
    spend_today: int
    total_balance: int


class Account:
    """Class for managing individual bank accounts"""

    def __init__(
        self,
        json: _MonzoAccountInfo,
        monzo_client: MonzoClient,
        balance_update_threshold: int = 15,
    ):
        self.json = json
        self._monzo_client = monzo_client

        self._balance_variables: _BalanceVariablesInfo

        self.last_balance_update = datetime(1970, 1, 1)
        self.balance_update_threshold = balance_update_threshold

    def _get_balance_property(
        self,
        var_name: Literal[
            "balance",
            "balance_including_flexible_savings",
            "currency",
            "local_currency",
            "local_exchange_rate",
            "local_spend",
            "spend_today",
            "total_balance",
        ],
    ) -> str | float | dict[str, str | int] | None:
        """Gets a value for a balance-specific property, updating the values if
         necessary (i.e. if they don't already exist). This also has a check to see if
         property is relevant for the given entity type and if not it just returns None

        Args:
            var_name (str): the name of the variable

        Returns:
            str: the value of the balance property
        """

        if (
            hasattr(self, "_balance_variables")
            and var_name not in self._balance_variables
        ):
            # Assume it's not a valid value so there's no point polling the API again
            return None

        if self.last_balance_update <= (
            datetime.utcnow() - timedelta(minutes=self.balance_update_threshold)
        ):
            self.update_balance_variables()

        return self._balance_variables[var_name]

    def update_balance_variables(self) -> None:
        """Updates the balance-related instance attributes with the latest values from
        the API
        """
        # pylint: disable=line-too-long
        self._balance_variables = self._monzo_client.get_json_response(  # type: ignore[assignment]
            "/balance", params={"account_id": self.id}
        )

        # convert values from pence to pounds
        for key, value in self._balance_variables.items():
            if isinstance(value, (int, float)):
                # pylint: disable=line-too-long
                self._balance_variables[key] = value / 100  # type: ignore[literal-required]

        self.last_balance_update = datetime.utcnow()

    @property
    def account_number(self) -> str | None:
        """
        Returns:
            str: the account's account number
        """
        return self.json.get("account_number")

    @property
    def balance(self) -> float | None:
        """
        Returns:
            float: the currently available balance of the account
        """
        return self._get_balance_property("balance")  # type: ignore[return-value]

    @property
    def balance_including_flexible_savings(self) -> float | None:
        """
        Returns:
            float: the currently available balance of the account, including flexible
             savings pots
        """
        return self._get_balance_property(
            "balance_including_flexible_savings"
        )  # type: ignore[return-value]

    @property
    def created_datetime(self) -> datetime | None:
        """
        Returns:
            datetime: when the account was created
        """
        if "created" not in self.json:
            return None

        return datetime.strptime(self.json["created"], DATETIME_FORMAT)

    @property
    def description(self) -> str | None:
        """
        Returns:
            str: the description of the account
        """
        return self.json.get("description")

    @property
    def id(self) -> str | None:
        """
        Returns:
            str: the account's UUID
        """
        return self.json["id"]

    @property
    def sort_code(self) -> str | None:
        """
        Returns:
            str: the account's sort code
        """
        return self.json.get("sort_code")

    @property
    def spend_today(self) -> float | None:
        """
        Returns:
            float: the amount spent from this account today (considered from approx
             4am onwards)
        """
        return self._get_balance_property("spend_today")  # type: ignore[return-value]

    @property
    def total_balance(self) -> str | None:
        """
        Returns:
            str: the sum of the currently available balance of the account and the
             combined total of all the user’s pots
        """
        return self._get_balance_property("total_balance")  # type: ignore[return-value]


class Pot:
    """Read-only class for Monzo pots"""

    def __init__(self, json: _MonzoPotInfo):
        self.json = json

    @property
    def available_for_bills(self) -> bool | None:
        """
        Returns:
            bool: if the pot can be used directly for bills
        """
        return self.json.get("available_for_bills")

    @property
    def balance(self) -> float:
        """
        Returns:
            float: the pot's balance in GBP
        """
        return float(self.json.get("balance", 0)) / 100

    @property
    def charity_id(self) -> str | None:
        """
        Returns:
            str: not sure!
        """
        return self.json.get("charity_id")

    @property
    def cover_image_url(self) -> str | None:
        """
        Returns:
            str: URL for the cover image
        """
        return self.json.get("cover_image_url")

    @property
    def created_datetime(self) -> datetime | None:
        """
        Returns:
            datetime: when the pot was created
        """
        if "created" not in self.json:
            return None

        return datetime.strptime(self.json["created"], DATETIME_FORMAT)

    @property
    def currency(self) -> str | None:
        """
        Returns:
            str: the currency of the pot
        """
        return self.json.get("currency")

    @property
    def current_account_id(self) -> str | None:
        """
        Returns:
            str: the UUID of the parent account
        """
        return self.json.get("current_account_id")

    @property
    def deleted(self) -> bool | None:
        """
        Returns:
            bool: has the pot been deleted
        """
        return self.json.get("deleted")

    @property
    def goal_amount(self) -> float | None:
        """
        Returns:
            float: the user-set goal amount for the pot
        """
        if not (goal_amount := self.json.get("goal_amount")):
            return None

        return goal_amount / 100

    @property
    def has_virtual_cards(self) -> bool | None:
        """
        Returns:
            bool: if the pot has virtual cards attached
        """
        return self.json.get("has_virtual_cards")

    @property
    def id(self) -> str:
        """
        Returns:
            str: the pot's UUID
        """
        return self.json["id"]

    @property
    def is_tax_pot(self) -> bool | None:
        """
        Returns:
            bool: is the pot taxed? I'm not sure
        """
        return self.json.get("is_tax_pot")

    @property
    def isa_wrapper(self) -> str | None:
        """
        Returns:
            str: is the pot ISA-wrapped?
        """
        return self.json.get("isa_wrapper")

    @property
    def locked(self) -> bool | None:
        """
        Returns:
            bool: is the pot locked
        """
        return self.json.get("locked")

    @property
    def name(self) -> str:
        """
        Returns:
            str: the name of the pot
        """
        return self.json["name"]

    @property
    def product_id(self) -> str | None:
        """
        Returns:
            str: the ID of the product applied to the pot (e.g. savings)
        """
        return self.json.get("product_id")

    @property
    def round_up(self) -> bool | None:
        """
        Returns:
            bool: is the pot where all round ups go
        """
        return self.json.get("round_up")

    @property
    def round_up_multiplier(self) -> float | None:
        """
        Returns:
            float: the multiplier applied to the pot's round-ups
        """
        return self.json.get("round_up_multiplier")

    @property
    def style(self) -> str | None:
        """
        Returns:
            str: the pot background image
        """
        return self.json.get("style")

    @property
    def type(self) -> str | None:
        """
        Returns:
            str: the type of pot (e.g. flex saver)
        """
        return self.json.get("type")

    @property
    def updated_datetime(self) -> datetime | None:
        """
        Returns:
            datetime: when the pot was updated last
        """
        if "updated" not in self.json:
            return None

        return datetime.strptime(self.json["updated"], DATETIME_FORMAT)

    def __str__(self) -> str:
        return f"{self.name} | {self.id}"


class MonzoClient(OauthClient):
    """Custom client for interacting with Monzo's API"""

    ACCESS_TOKEN_ENDPOINT = "https://api.monzo.com/oauth2/token"
    BASE_URL = "https://api.monzo.com"
    CREDS_FILE_PATH = user_data_dir(file_name="monzo_api_creds.json")

    def __init__(
        self,
        *,
        client_id: str,
        client_secret: str,
        redirect_uri: str = "http://0.0.0.0:5001/get_auth_code",
        access_token_expiry_threshold: int = 60,
        log_requests: bool = False,
        creds_cache_path: str | None = None,
    ):
        super().__init__(
            client_id=client_id,
            client_secret=client_secret,
            base_url=self.BASE_URL,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
            redirect_uri=redirect_uri,
            access_token_expiry_threshold=access_token_expiry_threshold,
            log_requests=log_requests,
            creds_cache_path=creds_cache_path or self.CREDS_FILE_PATH,
        )

        self._current_account: Account | None = None

    def deposit_into_pot(
        self, pot: Pot, amount_pence: int, dedupe_id: str | None = None
    ) -> None:
        """Move money from an account owned by the currently authorised user into one
         of their pots

        Args:
            pot (Pot): the target pot
            amount_pence (int): the amount of money to deposit, in pence
            dedupe_id (str): unique string used to de-duplicate deposits. Will be
             created if not provided
        """

        dedupe_id = dedupe_id or "|".join([pot.id, str(amount_pence)])

        res = put(
            f"{self.BASE_URL}/pots/{pot.id}/deposit",
            headers=self.request_headers,
            data={
                "source_account_id": self.current_account.id,
                "amount": amount_pence,
                "dedupe_id": dedupe_id,
            },
        )
        res.raise_for_status()

    def list_accounts(
        self, ignore_closed: bool = True, account_type: str | None = None
    ) -> Generator[Account, None, None]:
        """Gets a list of the user's accounts

        Args:
            ignore_closed (bool): whether to include closed accounts in the response
            account_type (str): the type of account(s) to find; submitted as param in
             request

        Yields:
            Account: Account instances, containing all related info
        """

        res = self.get_json_response(
            "/accounts", params={"account_type": account_type} if account_type else None
        )

        for account in res.get("accounts", []):
            if ignore_closed and account.get("closed", False) is True:
                continue
            yield Account(account, self)

    def list_pots(self, ignore_deleted: bool = True) -> Generator[Pot, None, None]:
        """Gets a list of the user's pots

        Args:
            ignore_deleted (bool): whether to include deleted pots in the response

        Yields:
            Pot: Pot instances, containing all related info
        """

        res = self.get_json_response(
            "/pots", params={"current_account_id": self.current_account.id}
        )

        for pot in res.get("pots", []):
            if ignore_deleted and pot.get("deleted", False) is True:
                continue
            yield Pot(pot)

    def get_pot_by_id(self, pot_id: str) -> Pot | None:
        """Get a pot from its ID

        Args:
            pot_id (str): the ID of the pot to find

        Returns:
            Pot: the Pot instance
        """
        for pot in self.list_pots():
            if pot.id == pot_id:
                return pot

        return None

    def get_pot_by_name(self, pot_name: str, exact_match: bool = False) -> Pot | None:
        """Get a pot from its name

        Args:
            pot_name (str): the name of the pot to find
            exact_match (bool): if False, all pot names will be cleansed before
             evaluation

        Returns:
            Pot: the Pot instance
        """
        if not exact_match:
            pot_name = cleanse_string(pot_name)

        for pot in self.list_pots():
            found_name = pot_name if exact_match else cleanse_string(pot.name).lower()
            if found_name.lower() == pot_name.lower():
                return pot

        return None

    @property
    def access_token_has_expired(self) -> bool:
        """Custom expiry check for Monzo client as JWT doesn't seem to include expiry
         time. Any errors/missing credentials result in a default value of True

        Returns:
            bool: expiry status of JWT
        """
        self._load_local_credentials()

        if not (access_token := self._credentials.get("access_token", False)):
            return True

        res = get(
            f"{self.BASE_URL}/ping/whoami",
            headers={"Authorization": f"Bearer {access_token}"},
        )

        return res.json().get("authenticated", False) is False

    @property
    def current_account(self) -> Account:
        """Get the main account for the Monzo user. We assume there'll only be one
         main account per user

        Returns:
            Account: the user's main account, instantiated
        """
        if not self._current_account:
            self._current_account = list(self.list_accounts(account_type="uk_retail"))[
                0
            ]

        return self._current_account



"""Helper functions specifically for parsing/manipulating XML"""


from logging import DEBUG, getLogger
from typing import Dict, Optional

from lxml import etree

LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)


def get_nsmap(
    *,
    root: Optional[etree._Element] = None,
    xml_doc: Optional[str] = None,
    warn_on_defaults: bool = False,
) -> Dict[str, str]:
    """Get the namespace map for an XML document

    Args:
        root (Element): an lxml Element from an XML document
        xml_doc (str): a raw XML document
        warn_on_defaults (bool): log a warning when an empty prefix is found and
         converted to a default value

    Returns:
        dict: a namespace mapping for the provided XML

    Raises:
        ValueError: if neither argument is provided
    """
    if not (root or xml_doc):
        raise ValueError("One of `root` or `xml_doc` should be non-null")

    root = root or etree.fromstring(xml_doc.encode())  # type: ignore
    nsmap = {}

    default_count = 0
    processed_urls = set()

    prefix: str
    url: str
    for prefix, url in root.xpath("//namespace::*"):  # type: ignore
        if url in processed_urls:
            continue

        if prefix is None:
            prefix = f"default_{default_count}"
            default_count += 1
            if warn_on_defaults:
                LOGGER.warning(
                    "Adding namespace url `%s` with prefix key `%s`", url, prefix
                )

        nsmap[prefix] = url
        processed_urls.add(url)

    return nsmap
